﻿
XIncludeFile("./FontWizzard_BF.pbi")

;==================================
; Module:          Rabbit
; Author:          Wilbert
; Date:            Nov 12, 2020
;==================================

DeclareModule Rabbit
  
  ;-RabbitCtx structure; Rabbit 
  Structure RabbitCtx
    X.l[8]  ; State
    C.l[8]  ; Counters
    Carry.l ; Carry for counters
  EndStructure
  
  ;-Exported functions
  Declare SetKeyAndIV(*Ctx.RabbitCtx, *Key, *IV)
  Declare ProcessBytes(*Ctx.RabbitCtx, *Input, *Output, Size)
  Declare CryptMemory(*Input, *Output, Size, *Key, *InitVector = #Null)
  
EndDeclareModule

Module Rabbit
  DisableDebugger
  EnableExplicit
  EnableASM
  
  ;- Private structures
  
  Structure Mem64Byte
    l.l[16]
  EndStructure
  
  ;- Private macros
  
  CompilerIf #PB_Compiler_Processor = #PB_Processor_x86
    Macro rbx : ebx : EndMacro  ; replace rbx with ebx on x86
    Macro rcx : ecx : EndMacro  ; replace rcx with ecx on x86
    Macro rdx : edx : EndMacro  ; replace rdx with edx on x86
    Macro rsi : esi : EndMacro  ; replace rsi with esi on x86
    Macro rdi : edi : EndMacro  ; replace rdi with edi on x86
    Macro rsp : esp : EndMacro  ; replace rsp with esp on x86
  CompilerEndIf 
  
  Macro M_AddWithCarry(n, v)
    mov eax, [rcx + n*4 + 32]   ; get c[n]
    adc eax, v                  ; add with carry v
    mov [rcx + n*4 + 32], eax   ; set c[n]
  EndMacro
  
  Macro M_GFunction(n)
    mov eax, [rcx + n*4]        ; get x[n]
    add eax, [rcx + n*4 + 32]   ; add c[n]
    mul eax                     ; eax*eax, result edx:eax
    XOr eax, edx                ; xor lower and upper result bits
    mov [rsp + n*4], eax        ; set g[n]
  EndMacro 
  
  Macro M_NewStateValue(n, n1, r1, n2, r2)
    mov eax, [rsp + n1*4]       ; get g[n1]
    mov edx, [rsp + n2*4]       ; get g[n2]
    rol eax, r1                 ; rol n1, r1
    CompilerIf r2
      rol edx, r2               ; rol n2, r2
    CompilerEndIf
    add eax, edx                ; add together
    add eax, [rsp + n*4]        ; add g[n]
    mov [rcx + n*4], eax        ; set x[n]
  EndMacro
  
  Macro M_SetStateAndCounter(k, n1, n2)
    CompilerIf k = 14           ; get key bits
      mov eax, [rdx]
      shl eax, 16
      Or ax, [rdx + 14]
    CompilerElse
      mov eax, [rdx + k]
    CompilerEndIf
    mov [rcx + n1*4], eax       ; set x[n1]
    rol eax, 16
    mov [rcx + n2*4 + 32], eax  ; set c[n2]
  EndMacro
  
  Macro M_ModifyCounter(n1, n2)
    mov eax, [rcx + n2*4]       ; get x[n2]
    XOr [rcx + n1*4 + 32], eax  ; c[n1] ! x[n2]
  EndMacro
  
  Macro M_XorStreamBytes(n1, n2, n3, n4, f=0)
    mov eax, [rcx + n3*4]       ; get x[n3]
    mov edx, [rcx + n4*4]       ; get x[n4]
    shrd eax, edx, 16           ; (x[n3] >> 16) | (x[n4] << 16)
    XOr eax, [rcx + n2*4]       ; xor x[n2]
    CompilerIf f=0
      XOr eax, [rsi + n1*4]
      mov [rdi + n1*4], eax
    CompilerElse
      mov [rsp + n1*4], eax
    CompilerEndIf
  EndMacro
  
  ;- Public functions
  
  Procedure SetKeyAndIV(*Ctx.RabbitCtx, *Key, *IV)
    Protected M.Mem64Byte
    
    ; *Key : Pointer to 128 bit key
    ; *IV  : Pointer to 64 bit initialization vector
    
    mov rcx, [p.p_Ctx]
    mov rdx, [p.p_Key]
    
    ; Generate initial state and counter values
    M_SetStateAndCounter( 0, 0, 4)
    M_SetStateAndCounter( 2, 5, 1)
    M_SetStateAndCounter( 4, 2, 6)
    M_SetStateAndCounter( 6, 7, 3)
    M_SetStateAndCounter( 8, 4, 0)
    M_SetStateAndCounter(10, 1, 5)
    M_SetStateAndCounter(12, 6, 2)
    M_SetStateAndCounter(14, 3, 7)
    
    ; Reset carry flag
    mov dword [rcx + 64], 0
    
    ; Process 64 bytes
    ProcessBytes(*Ctx, @M, @M, 64)
    
    ; Modify counters
    mov rcx, [p.p_Ctx]
    M_ModifyCounter(0, 4)
    M_ModifyCounter(1, 5)
    M_ModifyCounter(2, 6)
    M_ModifyCounter(3, 7)
    M_ModifyCounter(4, 0)
    M_ModifyCounter(5, 1)
    M_ModifyCounter(6, 2)
    M_ModifyCounter(7, 3)
    
    If *IV
      ; Modify counter values
      mov rcx, [p.p_Ctx]
      mov rdx, [p.p_IV]
      mov eax, [rdx]
      mov edx, [rdx + 4]
      XOr [rcx + 32], eax
      XOr [rcx + 40], edx
      XOr [rcx + 48], eax
      XOr [rcx + 56], edx
      rol edx, 16
      xchg ax, dx
      rol eax, 16
      XOr [rcx + 36], eax
      XOr [rcx + 44], edx
      XOr [rcx + 52], eax
      XOr [rcx + 60], edx
      ; Process 64 bytes
      ProcessBytes(*Ctx, @M, @M, 64)
    EndIf
    
  EndProcedure
  
  Procedure ProcessBytes(*Ctx.RabbitCtx, *Input, *Output, Size)
    Protected.i reg_bx, reg_si, reg_di
    If Size
      mov [p.v_reg_bx], rbx
      mov [p.v_reg_si], rsi
      mov [p.v_reg_di], rdi
      mov rcx, [p.p_Ctx]    ; rcx = *Ctx
      mov rsi, [p.p_Input]  ; rsi = *Input
      mov rdi, [p.p_Output] ; rdi = *Output
      mov rbx, [p.v_Size]   ; rbx = Size
      sub rsp, 32           ; alloc g space
      !.loop16b:
      
      ; Calculate new counter values
      mov edx, [rcx + 64] ; load carry
      shr edx, 1
      M_AddWithCarry(0, 0x4d34d34d)
      M_AddWithCarry(1, 0xd34d34d3)
      M_AddWithCarry(2, 0x34d34d34)
      M_AddWithCarry(3, 0x4d34d34d)
      M_AddWithCarry(4, 0xd34d34d3)
      M_AddWithCarry(5, 0x34d34d34)
      M_AddWithCarry(6, 0x4d34d34d)
      M_AddWithCarry(7, 0xd34d34d3)
      adc edx, edx
      mov [rcx + 64], edx ; save carry
      
      ; Calculate the g-functions
      M_GFunction(0)
      M_GFunction(1)
      M_GFunction(2)
      M_GFunction(3)
      M_GFunction(4)
      M_GFunction(5)
      M_GFunction(6)
      M_GFunction(7)
      
      ; Calculate new state values
      M_NewStateValue(0, 7, 16, 6, 16)
      M_NewStateValue(1, 0,  8, 7,  0)
      M_NewStateValue(2, 1, 16, 0, 16)
      M_NewStateValue(3, 2,  8, 1,  0)
      M_NewStateValue(4, 3, 16, 2, 16)
      M_NewStateValue(5, 4,  8, 3,  0)
      M_NewStateValue(6, 5, 16, 4, 16)
      M_NewStateValue(7, 6,  8, 5,  0)
      
      ; Xor stream bytes
      sub rbx, 16
      !jc .final
      M_XorStreamBytes(0, 0, 5, 3)
      M_XorStreamBytes(1, 2, 7, 5)
      M_XorStreamBytes(2, 4, 1, 7)
      M_XorStreamBytes(3, 6, 3, 1)
      add rsi, 16
      add rdi, 16
      test rbx, rbx
      !jnz .loop16b
      !jmp .done
      !.final:
      M_XorStreamBytes(0, 0, 5, 3, 1)
      M_XorStreamBytes(1, 2, 7, 5, 1)
      M_XorStreamBytes(2, 4, 1, 7, 1)
      M_XorStreamBytes(3, 6, 3, 1, 1)
      add rbx, 15
      !.loop1b:
      movzx eax, byte [rsi + rbx]
      XOr al, [rsp + rbx]
      mov [rdi + rbx], al
      sub rbx, 1
      !jnc .loop1b
      !.done:
      
      add rsp, 32
      mov rbx, [p.v_reg_bx]
      mov rsi, [p.v_reg_si]
      mov rdi, [p.v_reg_di]
    EndIf
  EndProcedure
  
  
  Procedure CryptMemory(*Input, *Output, Size, *Key, *InitVector = #Null)
    ; Encrypt/decrypt a block of memory
    ; ---------------------------------   
    ; *Key : Pointer to 128 bit key
    ; *IV  : Pointer to 64 bit initialization vector   
    Protected Ctx.RabbitCtx
    SetKeyAndIV(@Ctx, *Key, *InitVector)
    ProcessBytes(@Ctx, *Input, *Output, Size)
  EndProcedure
  
EndModule

;##############################################################

; Saki's Multi module Rabbit QAES Crypter

; Rabbit based three-stage special batch drag 'n' drop file crypter with CBC mode based IV handling

; Font_Wizzard_BF Module needed
; https://www.purebasic.fr/english/viewtopic.php?f=12&t=76338

; Wilberts Rabbit Crypter module needed
; https://www.purebasic.fr/english/viewtopic.php?f=27&t=76197&start=30

; This crypter combines several modules.
; The very cool Rabbit crypter Module by Wilbert.
; An amazingly working Progressbar_EX Module by walbus.
; An Drag 'n' Drop full featured module by walbus.
; The Rabbit QAES File Crypter Module by Saki.
; The required work code by Saki.

Define KeyFile$=GetHomeDirectory()+"Sakis_3Rabbit_KeyFile" ; Define here the KeyFile

; ==== Progressbar_EX ====
DeclareModule Progressbar_EX
  Declare Progressbar_EX(progressbar_ID, flags, foreground_color=$32CD32, background_color=$DDDDDD)
EndDeclareModule

Module Progressbar_EX
  ; This is a module for replacing a OS progressbar with a multi color progressbar
  ; Author : Werner Albus
  
  ; Add simple the Progressbar_EX call after any progressbar SetGadgetState calls
  ; SetGadgetState(progressbar_ID, 100*writen/file_size) ; OS progress bar
  ; Progressbar_EX(progressbar_ID, flags, $32CD32, $FFFF) ; Progressbar_EX
  
  ; This module can handle one progressbar in a event loop
  ; For handling different progressbars at the same loop, create other modules
  ; As sample, Progressbar_EX_1, Progressbar_EX_2, for each different progressbar one
  
  Procedure Progressbar_EX(progressbar_ID, flags, foreground_color=$32CD32, background_color=$DDDDDD)
    ; Available flags : #PB_ProgressBar_Vertical
    ; -1 remove the Progressbar_EX
    Static image_ID, image_gadget_ID
    If flags=-1
      If IsImage(image_ID) : FreeImage(image_ID) : image_ID=0 : EndIf
      If IsGadget(image_gadget_ID) : FreeGadget(image_gadget_ID) : image_gadget_ID=0 : EndIf
      DisableGadget(progressbar_ID, 0)
      HideGadget(progressbar_ID, 0)
      ProcedureReturn 0
    EndIf
    Protected state=GetGadgetState(progressbar_ID)
    Protected width.f=GadgetWidth(progressbar_ID)*DesktopResolutionX()
    Protected height.f=GadgetHeight(progressbar_ID)*DesktopResolutionY()
    Protected min=GetGadgetAttribute(progressbar_ID, #PB_ProgressBar_Minimum)
    Protected max=GetGadgetAttribute(progressbar_ID, #PB_ProgressBar_Maximum)
    If min<1 : min=1 : EndIf : If max<1 : max=1 : EndIf
    DisableGadget(progressbar_ID, 1)
    HideGadget(progressbar_ID, 1)
    If Not image_ID : image_ID=CreateImage(#PB_Any, width, height, 24, background_color) : EndIf
    StartDrawing(ImageOutput(image_ID))
    Box(0, 0, width, height, background_color)
    If flags=#PB_ProgressBar_Vertical
      Box(1, height-(state-min)*(height/(max-min)), width-1, height, foreground_color)
    Else
      Box(0, 1, (state-min)*(width/(max-min)), height-1, foreground_color)
    EndIf
    DrawingMode(#PB_2DDrawing_Outlined)
    Box(0, 0, width, height, $BBBBBB)
    StopDrawing()
    If Not image_gadget_ID
      image_gadget_ID=ImageGadget(#PB_Any, GadgetX(progressbar_ID), GadgetY(progressbar_ID), width, height, ImageID(image_ID))
    EndIf
    SetGadgetState(image_gadget_ID, ImageID(image_ID))
    ProcedureReturn image_gadget_ID
  EndProcedure
EndModule
UseModule Progressbar_EX

; #################### QAES Rabbit File Crypter ###########################

; ========= QAES CBC mode file crypter module with ProgressBar for Rabbit ==========

; Rabbit based three-stage special batch drag 'n' drop file crypter with CBC mode based IV handling

; By Saki - Team GFX_Wizzard_BF, QAES
; No warranty whatsever, using on your own risk
; Use Wilberts Rabbit module
; https://www.purebasic.fr/english/viewtopic.php?f=27&t=76197
; Each encrypted file is absolutelly different, also files with the same content are ever different
; Automatic check of encrypted files for integrity when decrypting - MD5 is presetted, you can simple change

DeclareModule QAES_Rabbit_FileCrypter
  UseModule Rabbit
  UseModule Progressbar_EX
  EnableExplicit
  ; Automatically Encrypting and Decrypting - This Crypter needs a crypt extender for encrypted files !
  Declare.s QAES_Rabbit_FileCrypter(path$, key_string$, integrity_check=1, progressbar_ID=-1)
  
  ; Encrypting
  Declare.s QAES_Rabbit_FileEncrypter(path$, key_string$, integrity_check=1, progressbar_ID=-1)
  
  ; Decrypting
  Declare.s QAES_Rabbit_FileDecrypter(path$, key_string$, integrity_check=1, progressbar_ID=-1)
  
  ; Integrity check a encrypted file
  Declare.s QAES_Rabbit_CryptedFileCheck(path$, key_string$, progressbar_ID=-1)
  
  ; Set a crypt extender - Needed for the automatic Encrypter-Decrypter function "QAES_Rabbit_FileCrypter()"
  Declare QAES_Rabbit_SetCryptExtender(crypt_extender_$=" [encrypted]")
  
EndDeclareModule

Module QAES_Rabbit_FileCrypter
  
  Global crypt_extender$=" [encrypted]"
  Global from_encrypter, from_decrypter
  
  Procedure.s QAES_Rabbit_FileCrypter(path$, key_string$, integrity_check=1, progressbar_ID=-1)
    
    Protected UseFileFingerprint=2 ; Look below - Presetted is md5 - This is fast and here as pre image using ok
    
    Static init, hash_byte_length, fingerprint
    If Not init
      Select UseFileFingerprint
        Case 1
          UseCRC32Fingerprint() : hash_byte_length=4 : fingerprint=#PB_Cipher_CRC32
        Case 2
          UseMD5Fingerprint() : hash_byte_length=16 : fingerprint=#PB_Cipher_MD5
        Case 3
          UseSHA1Fingerprint() : hash_byte_length=20 : fingerprint=#PB_Cipher_SHA1
        Case 4
          UseSHA3Fingerprint() : hash_byte_length=32 : fingerprint=#PB_Cipher_SHA3
      EndSelect
      init=OpenCryptRandom()
    EndIf
    If Not init
      ProcedureReturn "ERROR - Open Crypt random fail"
    EndIf
    
    If from_decrypter=2
      Protected cryptfile_check=1
    EndIf
    If cryptfile_check : integrity_check=1 : EndIf
    
    If Not (from_encrypter+from_decrypter+cryptfile_check) And crypt_extender$=""
      ProcedureReturn "ERROR - The automatic crypter needs a crypt extender"
    EndIf
    
    Protected crypt_marker_0.q=-1 ; Crypt marker - without integrity check
    Protected crypt_marker_1.q=-2 ; Crypt marker - with integrity check
    
    Protected Dim key.q(3)
    Protected *buffer
    Protected result$, file_fingerprint$
    Protected iv.q, iv_.q, iv_1.q, crypt_marker_check.q
    Protected fingerprint_ID, ii, iii, file, file_size, readed, writen, equal, block_size
    Protected save_hash_length, encrypted, timer, progressbar, amount, amount_1, extender_length=16
    
    If FileSize(path$)<0
      ProcedureReturn "ERROR - File not found"
    EndIf
    
    If key_string$=""
      ProcedureReturn "ERROR - Set firstly a key as string"
    EndIf
    
    If IsGadget(progressbar_ID)
      progressbar=#True
    EndIf
    
    If Not CryptRandomData(@iv, 8)
      ProcedureReturn "ERROR - Crypt random fail"
    EndIf
    
    key_string$+"i4380)/658909>290986`ßö.,;MDBCX?2\" ; Add a static salt - you can change
    
    key_string$=StringFingerprint(key_string$, #PB_Cipher_SHA3, 256, #PB_Ascii) ; Change here nothing
    If Len(key_string$)<>64
      ProcedureReturn "ERROR - Can not create the key"
    EndIf
    For ii = 0 To 31
      PokeA(@key(0)+ii, Val("$"+PeekS(@key_string$+iii, 2)))
      iii+SizeOf(character)<<1
    Next
    
    file=ReadFile(#PB_Any, path$)
    
    If Not file
      ProcedureReturn "ERROR - Can not read the file"
    EndIf
    
    If Lof(file)>15
      FileSeek(file, Lof(file)-8)
      crypt_marker_check=ReadQuad(file)
      FileSeek(file, Lof(file)-extender_length)
      iv_1=ReadQuad(file)
      CryptMemory(@crypt_marker_check, @crypt_marker_check, 8, @key(0), @iv_1)
      CryptMemory(@crypt_marker_check, @crypt_marker_check, 8, @key(2), @iv_1)
      CryptMemory(@crypt_marker_check, @crypt_marker_check, 8, @key(1), @iv_1)
      If crypt_marker_check=crypt_marker_0 Or crypt_marker_check=crypt_marker_1
        If from_encrypter
          CloseFile(file)
          ProcedureReturn "ERROR - This file is already encrypted"
        EndIf
        If from_decrypter And Not integrity_check
          save_hash_length=hash_byte_length
        Else
          integrity_check=Bool(crypt_marker_check=crypt_marker_1)
        EndIf
        encrypted=#True ; Crypt marker
        iv=iv_1
      Else
        If cryptfile_check Or from_decrypter
          CloseFile(file)
          #text_1=#LF$+"A encrypted file with password is needed"
          If from_decrypter
            ProcedureReturn "ERROR - Can not decrypt this file"+#text_1
          Else
            ProcedureReturn "ERROR - Can not check this file"+#text_1
          EndIf
        EndIf
        If Not from_encrypter
          If Right(path$, Len(crypt_extender$))=crypt_extender$
            CloseFile(file)
            ProcedureReturn "ERROR - Can not decrypt this file"
          EndIf
        EndIf
      EndIf
    Else
      If cryptfile_check Or from_decrypter
        CloseFile(file)
        ProcedureReturn "ERROR - This is not a encrypted file"
      EndIf
    EndIf
    CloseFile(file)
    
    If Not integrity_check
      hash_byte_length=0
    EndIf
    
    file=OpenFile(#PB_Any, path$)
    If Not file
      ProcedureReturn "ERROR - Can not open the file"
    EndIf
    
    timer=ElapsedMilliseconds()
    file_size=Lof(file)
    block_size=4096*4
    FileBuffersSize(file, block_size )
    *buffer=AllocateMemory(block_size+64)
    amount=file_size/block_size
    If Not *buffer
      CloseFile(file)
      ProcedureReturn "ERROR - Allocate memory fail"
    EndIf
    
    Macro Add_FingerprintBuffer(ID, buffer, bytes)
      If bytes>0
        AddFingerprintBuffer(ID, buffer, bytes)
      EndIf
    EndMacro
    If integrity_check
      fingerprint_ID=StartFingerprint(#PB_Any, fingerprint)
    EndIf
    Macro create_fingerprint
      file_fingerprint$=StringFingerprint(file_fingerprint$, fingerprint, 256, #PB_Ascii)
    EndMacro
    
    Macro get_all_events
      While WindowEvent() : Wend
    EndMacro   
    
    Repeat
      readed=ReadData(file, *buffer, block_size)
      If iv_
        If encrypted
          iv_1=PeekQ(*buffer)
          If readed
            If amount_1=amount
              CryptMemory(*buffer, *buffer, readed-extender_length-hash_byte_length, @key(0), @iv_)
              CryptMemory(*buffer, *buffer, readed-extender_length-hash_byte_length, @key(2), @iv_)
              CryptMemory(*buffer, *buffer, readed-extender_length-hash_byte_length, @key(1), @iv_)
              If integrity_check
                Add_FingerprintBuffer(fingerprint_ID, *buffer, readed-extender_length-hash_byte_length)
              EndIf
            Else
              CryptMemory(*buffer, *buffer, readed, @key(0), @iv_)
              CryptMemory(*buffer, *buffer, readed, @key(2), @iv_)
              CryptMemory(*buffer, *buffer, readed, @key(1), @iv_)
              If integrity_check
                Add_FingerprintBuffer(fingerprint_ID, *buffer, readed)
              EndIf
            EndIf
          EndIf
          iv_=iv_1
        Else
          If readed
            If integrity_check
              Add_FingerprintBuffer(fingerprint_ID, *buffer, readed)
            EndIf
            CryptMemory(*buffer, *buffer, readed, @key(0), @iv_)
            CryptMemory(*buffer, *buffer, readed, @key(2), @iv_)
            CryptMemory(*buffer, *buffer, readed, @key(1), @iv_)
          EndIf
          iv_=PeekQ(*buffer)
        EndIf
      Else
        If encrypted
          iv_=PeekQ(*buffer)
          If readed
            If amount_1=amount
              CryptMemory(*buffer, *buffer, readed-extender_length-hash_byte_length, @key(0), @iv)
              CryptMemory(*buffer, *buffer, readed-extender_length-hash_byte_length, @key(2), @iv)
              CryptMemory(*buffer, *buffer, readed-extender_length-hash_byte_length, @key(1), @iv)
              If integrity_check
                Add_FingerprintBuffer(fingerprint_ID, *buffer, readed-extender_length-hash_byte_length)
              EndIf
            Else
              CryptMemory(*buffer, *buffer, readed, @key(0), @iv)
              CryptMemory(*buffer, *buffer, readed, @key(2), @iv)
              CryptMemory(*buffer, *buffer, readed, @key(1), @iv)
              If integrity_check
                Add_FingerprintBuffer(fingerprint_ID, *buffer, readed)
              EndIf
            EndIf
          EndIf
        Else
          If readed
            If integrity_check
              Add_FingerprintBuffer(fingerprint_ID, *buffer, readed)
            EndIf
            CryptMemory(*buffer, *buffer, block_size, @key(0), @iv)
            CryptMemory(*buffer, *buffer, block_size, @key(2), @iv)
            CryptMemory(*buffer, *buffer, block_size, @key(1), @iv)
            iv_=PeekQ(*buffer)
          EndIf
        EndIf
      EndIf
      If cryptfile_check
        writen+readed
      Else
        FileSeek(file, -readed, #PB_Relative)
        writen+WriteData(file, *buffer, readed)
      EndIf
      If progressbar
        If ElapsedMilliseconds()>timer+30
          SetGadgetState(progressbar_ID, 100*writen/file_size)
          If encrypted
            If from_decrypter=2
              Progressbar_EX(progressbar_ID, 0, $FFBF00, $FFFFFF) ; Tweak the OS progressbar :)
            Else
              Progressbar_EX(progressbar_ID, 0, $00FF00, $FFFFFF) ; Tweak the OS progressbar :)
            EndIf
          Else
            Progressbar_EX(progressbar_ID, 0, $0000FF, $FFFFFF) ; Tweak the OS progressbar :)
          EndIf
          timer=ElapsedMilliseconds()
        EndIf
      EndIf
      amount_1+1
      get_all_events
    Until Not readed
    
    If integrity_check
      file_fingerprint$=FinishFingerprint(fingerprint_ID)
      create_fingerprint
      If Len(file_fingerprint$)<>hash_byte_length*2
        CloseFile(file)
        FreeMemory(*buffer)
        ProcedureReturn "ERROR - Can not create the file fingerprint"
      EndIf
      iii=0
      For ii = 0 To hash_byte_length
        PokeA(*buffer+ii, Val("$"+PeekS(@file_fingerprint$+iii, 2)))
        iii+SizeOf(character)<<1
      Next
      CryptMemory(*buffer, *buffer, hash_byte_length, @key(0), @iv)
      CryptMemory(*buffer, *buffer, hash_byte_length, @key(2), @iv)
      CryptMemory(*buffer, *buffer, hash_byte_length, @key(1), @iv)
    EndIf
    
    If integrity_check Or save_hash_length
      If encrypted
        If save_hash_length
          FileSeek(file, Lof(file)-extender_length-save_hash_length)
        Else
          FileSeek(file, Lof(file)-extender_length-hash_byte_length)
          ReadData(file, *buffer+hash_byte_length, hash_byte_length)
          FileSeek(file, Lof(file)-extender_length-hash_byte_length)
        EndIf
        If Not cryptfile_check
          TruncateFile(file)
        EndIf
        equal=CompareMemory(*buffer, *buffer+hash_byte_length, hash_byte_length)
      Else
        If Not cryptfile_check
          WriteData(file, *buffer, hash_byte_length)
          WriteQuad(file, iv)
          CryptMemory(@crypt_marker_1, @crypt_marker_1, 8, @key(0), @iv)
          CryptMemory(@crypt_marker_1, @crypt_marker_1, 8, @key(2), @iv)
          CryptMemory(@crypt_marker_1, @crypt_marker_1, 8, @key(1), @iv)
          WriteQuad(file, crypt_marker_1)
        EndIf
      EndIf
    Else
      If encrypted
        FileSeek(file, Lof(file)-extender_length)
        TruncateFile(file)
      Else
        WriteQuad(file, iv)
        CryptMemory(@crypt_marker_0, @crypt_marker_0, 8, @key(0), @iv)
        CryptMemory(@crypt_marker_0, @crypt_marker_0, 8, @key(2), @iv)
        CryptMemory(@crypt_marker_0, @crypt_marker_0, 8, @key(1), @iv)
        WriteQuad(file, crypt_marker_0)
      EndIf
    EndIf
    
    CloseFile(file)
    FreeMemory(*buffer)
    
    Macro set_progress
      If progressbar
        SetGadgetState(progressbar_ID, 100)
        If encrypted
          If from_decrypter=2
            Progressbar_EX(progressbar_ID, 0, $FFBF00, $FFFFFF) ; Tweak the OS progressbar :)
          Else
            Progressbar_EX(progressbar_ID, 0, $00FF00, $FFFFFF) ; Tweak the OS progressbar :)
          EndIf
        Else
          Progressbar_EX(progressbar_ID, 0, $0000FF, $FFFFFF) ; Tweak the OS progressbar :)
        EndIf
        get_all_events
      EndIf
    EndMacro
    
    If integrity_check And Not save_hash_length
      If encrypted And Not equal
        If Not cryptfile_check
          If Right(path$, Len(crypt_extender$))=crypt_extender$
            RenameFile(path$, Left(path$, Len(path$)-Len(crypt_extender$))+" [File Hash broken]")
          EndIf
        EndIf
        set_progress
        ProcedureReturn "ERROR - File Hash broken"
      EndIf
      If cryptfile_check
        set_progress
        ProcedureReturn "ALLok - File integrity has checked OK"
      EndIf
    EndIf
    
    If writen<>file_size
      RenameFile(path$, path$+" [Writen fails]")
      set_progress
      ProcedureReturn "ERROR - Writen fails"+#LF$+"Coded Bytes : "+Str(writen)
    EndIf
    
    If Not from_encrypter+from_decrypter
      If Right(path$, Len(crypt_extender$))=crypt_extender$
        RenameFile(path$, Left(path$, Len(path$)-Len(crypt_extender$)))
        set_progress
        If integrity_check
          ProcedureReturn "ALLok - File decrypted - Integrity checked OK"
        Else
          ProcedureReturn "ALLok - File decrypted"
        EndIf
      Else
        RenameFile(path$, path$+crypt_extender$)
        set_progress
        If integrity_check
          ProcedureReturn "ALLok - File encrypted - Integrity check added"
        Else
          ProcedureReturn "ALLok - File encrypted"
        EndIf
      EndIf
    EndIf
    
    If from_encrypter
      RenameFile(path$, path$+crypt_extender$)
      set_progress
      If integrity_check
        ProcedureReturn "ALLok - File encrypted - Integrity check added"
      Else
        ProcedureReturn "ALLok - File encrypted"
      EndIf
    EndIf
    
    If from_decrypter
      If Right(path$, Len(crypt_extender$))=crypt_extender$
        RenameFile(path$, Left(path$, Len(path$)-Len(crypt_extender$)))
      EndIf
      set_progress
      If integrity_check
        ProcedureReturn "ALLok - File decrypted - Integrity checked OK"
      Else
        ProcedureReturn "ALLok - File decrypted"
      EndIf
    EndIf
    
  EndProcedure
  
  Procedure.s QAES_Rabbit_FileEncrypter(path$, key_string$, integrity_check=1, progressbar_ID=-1)
    from_encrypter=1
    Protected result$=QAES_Rabbit_FileCrypter(path$, key_string$, integrity_check, progressbar_ID)
    from_encrypter=0
    ProcedureReturn result$
  EndProcedure
  
  Procedure.s QAES_Rabbit_FileDecrypter(path$, key_string$, integrity_check=1, progressbar_ID=-1)
    from_decrypter=1
    Protected result$=QAES_Rabbit_FileCrypter(path$, key_string$, integrity_check, progressbar_ID)
    from_decrypter=0
    ProcedureReturn result$
  EndProcedure
  
  Procedure.s QAES_Rabbit_CryptedFileCheck(path$, key_string$, progressbar_ID=-1)
    from_decrypter=2
    Protected result$=QAES_Rabbit_FileCrypter(path$, key_string$, 0, progressbar_ID)
    from_decrypter=0
    ProcedureReturn result$
  EndProcedure
  
  Procedure QAES_Rabbit_SetCryptExtender(crypt_extender_$=" [encrypted]")
    crypt_extender$=crypt_extender_$
    ProcedureReturn 1
  EndProcedure
  
EndModule

; === Module QAES get folders and files ===
DeclareModule QAES_get_folder_files
  Declare QAES_get_folder_files(mode, List PathList$(), path_found$, include_hidden=#True)
EndDeclareModule

Module QAES_get_folder_files
  ; Author : Werner Albus
  EnableExplicit
  
  Procedure QAES_get_folder_files(mode, List PathList$(), path_found$, include_hidden=#True)
    
    CompilerIf #PB_Compiler_OS=#PB_OS_Windows
      If Right(path_found$, 1)<>"\" : path_found$+"\" : EndIf
    CompilerElse
      If Right(path_found$, 1)<>"/" : path_found$+"/" : EndIf
    CompilerEndIf
    
    ; Set mode=1 for broken file hash messages
    Protected directory, entry_attributes, entry_type
    Protected entry_name$, result$
    
    directory=ExamineDirectory(#PB_Any, path_found$, "*.*")
    If directory
      While NextDirectoryEntry(directory)
        While WindowEvent() : Wend
        entry_name$=DirectoryEntryName(directory)
        entry_attributes=DirectoryEntryAttributes(directory)
        entry_type=DirectoryEntryType(directory)
        If entry_name$="." Or entry_name$=".." : Continue : EndIf
        If Not include_hidden
          CompilerIf #PB_Compiler_OS=#PB_OS_Windows
            If entry_attributes & #PB_FileSystem_Hidden : Continue : EndIf
          CompilerEndIf
          CompilerIf #PB_Compiler_OS=#PB_OS_Linux
            If Left(entry_name$, 1) = "." : Continue : EndIf
          CompilerEndIf
        EndIf
        Select entry_type
          Case #PB_DirectoryEntry_Directory
            CompilerIf #PB_Compiler_OS=#PB_OS_Windows
              QAES_get_folder_files(1, PathList$(), path_found$+entry_name$+"\", include_hidden)
            CompilerElse
              QAES_get_folder_files(1, PathList$(), path_found$+entry_name$+"/", include_hidden)
            CompilerEndIf
          Case #PB_DirectoryEntry_File
            AddElement(PathList$())
            PathList$()=path_found$+entry_name$
        EndSelect
      Wend
      FinishDirectory(directory)
      ProcedureReturn 1
    EndIf
  EndProcedure
EndModule
UseModule QAES_get_folder_files

; ====== Little UI for Drag and Drop file crypting ======
CompilerIf #PB_Compiler_IsMainFile
  
  ; === Use the QAES file crypter ===
  
  UseModule QAES_Rabbit_FileCrypter
  
  If Not OpenCryptRandom()
    MessageRequester(" ERROR", "OpenCryptRandom fail")
    End
  EndIf
  
  UseSHA3Fingerprint()
  
  EnableExplicit
  
  Define password$, password_1$, path$
  Define password_change, in_loop
  
  If #PB_Compiler_OS=#PB_OS_Linux And Not Subsystem("qt") And Not Subsystem("gtk2")
    MessageRequester(" Hint", "For best design use Subsystem qt or Subsystem gtk2")
  EndIf
  
  CompilerIf #PB_Compiler_OS=#PB_OS_Linux And Not Subsystem("gtk2") And Not Subsystem("qt")
    Define window_ID=OpenWindow(#PB_Any, 0, 0, 350, 270, "Saki's 3Rabbit file crypter",
                                #PB_Window_ScreenCentered|#PB_Window_SystemMenu|#PB_Window_MinimizeGadget|#PB_Window_Invisible)
  CompilerElse
    Define window_ID=OpenWindow(#PB_Any, 0, 0, 350, 255, "Saki's 3Rabbit file crypter",
                                #PB_Window_ScreenCentered|#PB_Window_SystemMenu|#PB_Window_MinimizeGadget|#PB_Window_Invisible)
  CompilerEndIf
  
  Define salt$="Soup without salt simply does not taste good - LOL"
  Define integrity_check=#True
  
  Define menu_ID=CreateMenu(#PB_Any, WindowID(window_ID))
  If menu_ID
    MenuTitle(" More ")
    MenuItem(1, "Change the Password")
    MenuBar()
    MenuItem(2, "Create and use a KeyFile in your UserDirectory")
    MenuBar()
    MenuItem(3, "Delete the KeyFile in your UserDirectory")
    MenuBar()
    MenuItem(4, "Import a Key from a external KeyFile")
    MenuBar()
    MenuItem(5, "Export a Key in a external KeyFile")
    MenuBar()
    MenuItem(6, "Delete a File")
    MenuBar()
    MenuItem(7, "Info")
    DisableMenuItem(menu_ID, 3, 1)
  EndIf
  
  Define button_encrypt_ID=ButtonImageGadget(#PB_Any, 10, 10, WindowWidth(window_ID)/2-10, 60, 0)
  Define button_decrypt_ID=ButtonImageGadget(#PB_Any, WindowWidth(window_ID)/2+10, 10, WindowWidth(window_ID)/2-20, 60, 0)
  Define button_integrity_check_ID=ButtonImageGadget(#PB_Any, 10, 80, WindowWidth(window_ID)-20, 60, 0)
  Define button_state_ID=ImageGadget(#PB_Any, 10, 150, WindowWidth(window_ID)-20, 40, 0)
  
  Define button_break_ID=ButtonImageGadget(#PB_Any, WindowWidth(window_ID)-65, 200, 55, 24, 0)
  
  EnableGadgetDrop(button_encrypt_ID, #PB_Drop_Files, #PB_Drag_Link)
  EnableGadgetDrop(button_decrypt_ID, #PB_Drop_Files, #PB_Drag_Link)
  EnableGadgetDrop(button_integrity_check_ID, #PB_Drop_Files, #PB_Drag_Link)
  
  Define progressbar_ID=ProgressBarGadget(#PB_Any, 10, 200,
                                          WindowWidth(window_ID)-80,
                                          24, 0, 100, #PB_ProgressBar_Smooth)
  
  DisableGadget(button_break_ID, 1)
  
  SetGadgetState(progressbar_ID, 0)
  Progressbar_EX(progressbar_ID, 0, 0, $FFFFFF) ; Tweak the OS progressbar :)
  
  PresetMaxFontsize_FW(12) ; Preset the max used font size
  
  ; TODO Font_Wizzard_BF enhancement
  Define text$="Encrypt"
  SetImageBasedGadgetText_FW(button_encrypt_ID,
                             "",            ; Font$, ; The necessary fonts are managed automatically
                             #PB_Font_Bold, ; font_flag,
                             #Black,        ; Text color
                             0,             ; Text adjustment - 0=center - 1=left - 2=right
                             $500000FF,     ; Background color - Ignore = -1
                             text$,         ; Text
                             0,             ; padding_x
                             0,             ; padding_y
                             1)             ; resize_factor.f
  
  ; TODO Font_Wizzard_BF enhancement
  Define text$="Decrypt"
  SetImageBasedGadgetText_FW(button_decrypt_ID,
                             "",            ; Font$, ; The necessary fonts are managed automatically
                             #PB_Font_Bold, ; font_flag,
                             #Black,        ; Text color
                             0,             ; Text adjustment - 0=center - 1=left - 2=right
                             $5000FF00,     ; Background color - Ignore = -1
                             text$,         ; Text
                             0,             ; padding_x
                             0,             ; padding_y
                             1)             ; resize_factor.f
  
  ; TODO Font_Wizzard_BF enhancement
  Define text$="Check encrypted files"
  SetImageBasedGadgetText_FW(button_integrity_check_ID,
                             "",            ; Font$, ; The necessary fonts are managed automatically
                             #PB_Font_Bold, ; font_flag,
                             #Black,        ; Text color
                             0,             ; Text adjustment - 0=center - 1=left - 2=right
                             $40FF0000,     ; Background color - Ignore = -1
                             text$,         ; Text
                             0,             ; padding_x
                             0,             ; padding_y
                             1)             ; resize_factor.f
  
  ; TODO Font_Wizzard_BF enhancement
  AddFrame_FW(1, $FF000000)
  Define text$="Drag 'n' Drop"+#LF$+"Multiple files and folders supported"
  SetImageBasedGadgetText_FW(button_state_ID,
                             "",            ; Font$, ; The necessary fonts are managed automatically
                             #PB_Font_Bold, ; font_flag,
                             #Black,        ; Text color
                             0,             ; Text adjustment - 0=center - 1=left - 2=right
                             -1,            ; Background color - Ignore = -1
                             text$,         ; Text
                             10,            ; padding_x
                             3,             ; padding_y
                             1)             ; resize_factor.f
  
  ; TODO Font_Wizzard_BF enhancement
  Define text$="Break"
  SetImageBasedGadgetText_FW(button_break_ID,
                             "",            ; Font$, ; The necessary fonts are managed automatically
                             #PB_Font_Bold, ; font_flag,
                             #Black,        ; Text color
                             0,             ; Text adjustment - 0=center - 1=left - 2=right
                             $9000FFFF,     ; Background color - Ignore = -1
                             text$,         ; Text
                             -5,            ; padding_x
                             0,             ; padding_y
                             1)             ; resize_factor.f=1
  
  
  NewList PathList$()
  
  #crypt_extender$=" [encrypted]"
  
  QAES_Rabbit_SetCryptExtender(#crypt_extender$)
  
  Procedure.s KeyStretching(key$, window_ID, progressbar_ID)
    ; Key stretching - By walbus
    #KeyStretchingLoops=6e5
    If IsWindow(window_ID)
      AddWindowTimer(window_ID, 1, 30)
    EndIf
    Define i
    For i=1 To #KeyStretchingLoops ; Iterations
      key$=StringFingerprint(key$, #PB_Cipher_SHA3, 512, #PB_Ascii)
      If IsWindow(window_ID)
        If WindowEvent()=#PB_Event_Timer And IsGadget(progressbar_ID)
          SetGadgetState(progressbar_ID, 100*i/#KeyStretchingLoops)
          Progressbar_EX(progressbar_ID, 100, $FF00FF, $FFFFFF) ; Tweak the OS progressbar :)
        EndIf
      EndIf
    Next i
    If IsWindow(window_ID) And IsGadget(progressbar_ID)
      SetGadgetState(progressbar_ID, 0)
      Progressbar_EX(progressbar_ID, 0, 0, $FFFFFF) ; Tweak the OS progressbar :)
    EndIf
    key$=StringFingerprint(key$, #PB_Cipher_SHA3, 256, #PB_Ascii) ; Finalization for Rabbit Dual Crypter
    ProcedureReturn key$
  EndProcedure
  
  If FileSize(keyFile$)=64
    Define file=ReadFile(#PB_Any, KeyFile$)
    If file
      password$=ReadString(file, #PB_Ascii)
      CloseFile(file)
    EndIf
  EndIf
  
  If Len(password$)=64
    ; TODO Font_Wizzard_BF enhancement
    PresetMaxFontsize_FW(11)
    AddFrame_FW(1, $FF000000)
    Define text$="KeyFile is loaded"
    SetImageBasedGadgetText_FW(button_state_ID,
                               "",            ; Font$, ; The necessary fonts are managed automatically
                               #PB_Font_Bold, ; font_flag,
                               #Black,        ; Text color
                               0,             ; Text adjustment - 0=center - 1=left - 2=right
                               -1,            ; Background color - Ignore = -1
                               text$,         ; Text
                               -5,            ; padding_x
                               0,             ; padding_y
                               1)             ; resize_factor.f
    PresetMaxFontsize_FW()
    
    If menu_ID
      DisableMenuItem(menu_ID, 3, 0)
    EndIf
    HideWindow(window_ID, 0)
  Else
    set_new_password:
    If password_change
      password_change=0
      HideWindow(window_ID, 1)
      password_1$=InputRequester("Set a new Password", "", "")
      If password_1$<>""
        Delay(300)
        HideWindow(window_ID, 0)
        password$=StringFingerprint(password_1$, #PB_Cipher_SHA3, 256, #PB_Ascii)
        password$=KeyStretching(password$, window_ID, progressbar_ID)
        ; TODO Font_Wizzard_BF enhancement
        PresetMaxFontsize_FW(11)
        AddFrame_FW(1, $FF000000)
        Define text$="Your Password is changed"
        SetImageBasedGadgetText_FW(button_state_ID,
                                   "",            ; Font$, ; The necessary fonts are managed automatically
                                   #PB_Font_Bold, ; font_flag,
                                   #Black,        ; Text color
                                   0,             ; Text adjustment - 0=center - 1=left - 2=right
                                   -1,            ; Background color - Ignore = -1
                                   text$,         ; Text
                                   -5,            ; padding_x
                                   0,             ; padding_y
                                   1)             ; resize_factor.f
        PresetMaxFontsize_FW()
      Else
        Delay(300)
        HideWindow(window_ID, 0)
      EndIf
    Else
      password$=InputRequester("Set firstly a Password", "", "")
      If password$="" : End : EndIf
      password$=StringFingerprint(password$, #PB_Cipher_SHA3, 256, #PB_Ascii)
      Delay(300)
      HideWindow(window_ID, 0)
      password$=KeyStretching(password$, window_ID, progressbar_ID)
    EndIf
    
    If Len(password$)<>64
      ; TODO Font_Wizzard_BF enhancement
      PresetMaxFontsize_FW(11)
      AddFrame_FW(1, $FF000000)
      Define text$="ERROR"+#LF$+"Can not create the key"
      SetImageBasedGadgetText_FW(button_state_ID,
                                 "",            ; Font$, ; The necessary fonts are managed automatically
                                 #PB_Font_Bold, ; font_flag,
                                 #Black,        ; Text color
                                 0,             ; Text adjustment - 0=center - 1=left - 2=right
                                 -1,            ; Background color - Ignore = -1
                                 text$,         ; Text
                                 -5,            ; padding_x
                                 0,             ; padding_y
                                 1)             ; resize_factor.f
      PresetMaxFontsize_FW()
      Delay(3000)
      End
    EndIf
  EndIf
  
  Define cryptfile_check
  
  Macro progress_1 ; #PB_Event_GadgetDrop
                   ; Define progress_foreground_color=$FFFFFF ; Tweak the OS progressbar :)
    path$=EventDropFiles()
    Define button_break_x=GadgetX(button_break_ID)*DesktopResolutionX()
    Define button_break_y=GadgetY(button_break_ID)*DesktopResolutionY()
    Define button_break_xx=(GadgetX(button_break_ID)+GadgetWidth(button_break_ID))*DesktopResolutionX()
    Define button_break_yy=(GadgetY(button_break_ID)+GadgetHeight(button_break_ID))*DesktopResolutionY()
    
    ; Get paths
    Repeat
      little_loop_1:
      win_event=WindowEvent()
      If WindowMouseX(window_ID)>button_break_x And WindowMouseY(window_ID)>button_break_y And
         WindowMouseX(window_ID)<button_break_xx And WindowMouseY(window_ID)<button_break_yy
        Delay(1)
        If win_event=#PB_Event_Gadget And EventGadget()=button_break_ID : Goto break_1_ : EndIf
        Goto little_loop_1_
      EndIf
      Define find_lf_0=FindString(#LF$+path$+#LF$, #LF$, find_lf_0)
      Define find_lf_1=FindString(#LF$+path$+#LF$, #LF$, find_lf_1)
      If find_lf_1
        Define result$=Mid(#LF$+path$+#LF$, find_lf_0+1, find_lf_1-find_lf_0-1)
        
        Define path_found$=result$
        
        If FileSize(path_found$)=-2 ; Folder
          QAES_get_folder_files(1, PathList$(), path_found$)
        Else
          If result$<>""
            AddElement(PathList$())
            PathList$()=result$
          EndIf
        EndIf
      EndIf
      If find_lf_1+3>find_lf_0
        find_lf_0=find_lf_1 : find_lf_1=find_lf_1+3
      Else
        Break
      EndIf
    ForEver
    find_lf_0=0 : find_lf_1=0
    
    ; Crypting
    Define result=0
    text$=""
    in_Loop=0
    
    ForEach PathList$()
      result+1
      little_loop_2:
      win_event=WindowEvent()
      If WindowMouseX(window_ID)>button_break_x And WindowMouseY(window_ID)>button_break_y And
         WindowMouseX(window_ID)<button_break_xx And WindowMouseY(window_ID)<button_break_yy
        
        If Not in_loop
          ; TODO Font_Wizzard_BF enhancement
          Define text$="Break"
          SetImageBasedGadgetText_FW(button_break_ID,
                                     "",            ; Font$, ; The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     $FF0000FF,     ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          in_loop=1
        EndIf
        
        If win_event=#PB_Event_Gadget And EventGadget()=button_break_ID
          DisableGadget(button_break_ID, 1)
          ; TODO Font_Wizzard_BF enhancement
          PresetMaxFontsize_FW(11)
          AddFrame_FW(1, $FF000000)
          Define text$="Break by user"+#CRLF$+"I will now undo the changes"
          SetImageBasedGadgetText_FW(button_state_ID,
                                     "",            ; Font$, ; The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     -1,            ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          PresetMaxFontsize_FW()
          ; TODO Font_Wizzard_BF enhancement
          Define text$="Break"
          SetImageBasedGadgetText_FW(button_break_ID,
                                     "",            ; Font$, ; The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     $9000FFFF,     ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          Delay(10)
          While WindowEvent() : Wend
          Delay(2000)
          Define result_1=result-1 : result=0
          ResetList(PathList$())
          ForEach PathList$()
            result+1
            result$=PathList$()
            If Right(result$, Len(#crypt_extender$))=#crypt_extender$
              result$=Left(result$, Len(result$)-Len(#crypt_extender$))
            Else
              result$+#crypt_extender$
            EndIf
            RenameFile(result$, result$)
            ; TODO Font_Wizzard_BF enhancement
            PresetMaxFontsize_FW(11)
            AddFrame_FW(1, $FF000000)
            Define text$=QAES_Rabbit_FileCrypter(result$, password$, integrity_check, progressbar_ID)
            SetImageBasedGadgetText_FW(button_state_ID,
                                       "",            ; Font$, ; The necessary fonts are managed automatically
                                       #PB_Font_Bold, ; font_flag,
                                       #Black,        ; Text color
                                       0,             ; Text adjustment - 0=center - 1=left - 2=right
                                       -1,            ; Background color - Ignore = -1
                                       text$,         ; Text
                                       -5,            ; padding_x
                                       0,             ; padding_y
                                       1)             ; resize_factor.f
            PresetMaxFontsize_FW()
            
            If result=result_1 : Goto break_1 : EndIf
          Next
        EndIf
        Goto little_loop_2
      EndIf
      ; TODO Font_Wizzard_BF enhancement
      If (GetGadgetData(button_encrypt_ID) And Right(GetFilePart(PathList$()), Len(#crypt_extender$))<>#crypt_extender$) Or
         (Not GetGadgetData(button_encrypt_ID) And Right(GetFilePart(PathList$()), Len(#crypt_extender$))=#crypt_extender$)
        Define text$=QAES_Rabbit_FileCrypter(PathList$(), password$, integrity_check, progressbar_ID)
      EndIf
      PresetMaxFontsize_FW(11)
      AddFrame_FW(1, $FF000000)
      SetImageBasedGadgetText_FW(button_state_ID,
                                 "",            ; Font$, ; The necessary fonts are managed automatically
                                 #PB_Font_Bold, ; font_flag,
                                 #Black,        ; Text color
                                 0,             ; Text adjustment - 0=center - 1=left - 2=right
                                 -1,            ; Background color - Ignore = -1
                                 text$,         ; Text
                                 -5,            ; padding_x
                                 0,             ; padding_y
                                 1)             ; resize_factor.f
      PresetMaxFontsize_FW()
    Next
    
    ; Crypting finish
    break_1:
    ClearList(PathList$())
    path$=""
  EndMacro
  
  Macro progress_2 ; #PB_Event_GadgetDrop
    path$=EventDropFiles()
    Define button_break_x=GadgetX(button_break_ID)*DesktopResolutionX()
    Define button_break_y=GadgetY(button_break_ID)*DesktopResolutionY()
    Define button_break_xx=(GadgetX(button_break_ID)+GadgetWidth(button_break_ID))*DesktopResolutionX()
    Define button_break_yy=(GadgetY(button_break_ID)+GadgetHeight(button_break_ID))*DesktopResolutionY()
    
    ; Get paths
    Repeat
      little_loop_1_:
      win_event=WindowEvent()
      If WindowMouseX(window_ID)>button_break_x And WindowMouseY(window_ID)>button_break_y And
         WindowMouseX(window_ID)<button_break_xx And WindowMouseY(window_ID)<button_break_yy
        Delay(1)
        If win_event=#PB_Event_Gadget And EventGadget()=button_break_ID
          Goto break_1_
        EndIf
        Goto little_loop_1_
      EndIf
      Define find_lf_0=FindString(#LF$+path$+#LF$, #LF$, find_lf_0)
      Define find_lf_1=FindString(#LF$+path$+#LF$, #LF$, find_lf_1)
      If find_lf_1
        Define result$=Mid(#LF$+path$+#LF$, find_lf_0+1, find_lf_1-find_lf_0-1)
        
        Define path_found$=result$
        
        If FileSize(path_found$)=-2 ; Folder
          QAES_get_folder_files(1, PathList$(), path_found$)
        Else
          If result$<>""
            AddElement(PathList$())
            PathList$()=result$
          EndIf
        EndIf
      EndIf
      If find_lf_1+3>find_lf_0
        find_lf_0=find_lf_1 : find_lf_1=find_lf_1+3
      Else
        Break
      EndIf
    ForEver
    find_lf_0=0 : find_lf_1=0
    
    ; Crypting
    Define result=0
    
    ForEach PathList$()
      result+1
      little_loop_2_:
      win_event=WindowEvent()
      If WindowMouseX(window_ID)>button_break_x And WindowMouseY(window_ID)>button_break_y And
         WindowMouseX(window_ID)<button_break_xx And WindowMouseY(window_ID)<button_break_yy
        If Not in_loop
          ; TODO Font_Wizzard_BF enhancement
          Define text$="Break"
          SetImageBasedGadgetText_FW(button_break_ID,
                                     "",            ; Font$, - The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     $FF0000FF,     ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          in_loop=1
        EndIf   
        Delay(10)
        If win_event=#PB_Event_Gadget And EventGadget()=button_break_ID
          
          PresetMaxFontsize_FW()
          ; TODO Font_Wizzard_BF enhancement
          Define text$="Break"
          SetImageBasedGadgetText_FW(button_break_ID,
                                     "",            ; Font$, - The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     $9000FFFF,     ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          Goto break_1_
        EndIf
        Goto little_loop_2_
      Else
        If in_loop
          ; TODO Font_Wizzard_BF enhancement
          Define text$="Break"
          SetImageBasedGadgetText_FW(button_break_ID,
                                     "",            ; Font$, - The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     $FF00FFFF,     ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          in_loop=0
        EndIf
      EndIf
      
      If cryptfile_check=1
        result$=QAES_Rabbit_CryptedFileCheck(PathList$(), password$, progressbar_ID)
      Else
        result$=QAES_Rabbit_FileCrypter(PathList$(), password$, integrity_check, progressbar_ID)
      EndIf
      ; TODO Font_Wizzard_BF enhancement
      PresetMaxFontsize_FW(11)
      AddFrame_FW(1, $FF000000)
      SetImageBasedGadgetText_FW(button_state_ID,
                                 "",            ; Font$, - The necessary fonts are managed automatically
                                 #PB_Font_Bold, ; font_flag,
                                 #Black,        ; Text color
                                 0,             ; Text adjustment - 0=center - 1=left - 2=right
                                 -1,            ; Background color - Ignore = -1
                                 result$,       ; Text
                                 -5,            ; padding_x
                                 0,             ; padding_y
                                 1)             ; resize_factor.f
      PresetMaxFontsize_FW()
      
      If Left(result$, 5)="ERROR"
        HideWindow(window_ID, 1)
        If MessageRequester("ERROR", PathList$()+#LF$+#LF$+result$+#LF$+#LF$+"Continue ?",
                            #PB_MessageRequester_YesNo)<>#PB_MessageRequester_Yes
          HideWindow(window_ID, 0)
          Goto break_1_
        EndIf
        HideWindow(window_ID, 0)
      EndIf
    Next
    
    ; Crypting finish
    break_1_:
    ClearList(PathList$())
    path$=""
  EndMacro
  
  Repeat
    Define win_event=WaitWindowEvent()
    If win_event=#PB_Event_Menu
      
      If EventMenu()=1
        password_change=1
        Goto set_new_password
      EndIf
      
      If EventMenu()=2
        QAES_Rabbit_SetCryptExtender("")
        QAES_Rabbit_FileEncrypter(KeyFile$, password$)
        QAES_Rabbit_SetCryptExtender()
        file=CreateFile(#PB_Any, KeyFile$)
        If IsFile(file)
          If WriteString(file, password$, #PB_Ascii)=Len(password$)
            CloseFile(file)
            ; TODO Font_Wizzard_BF enhancement
            PresetMaxFontsize_FW(11)
            AddFrame_FW(1, $FF000000)
            Define text$="KeyFile created"
            SetImageBasedGadgetText_FW(button_state_ID,
                                       "",            ; Font$, ; The necessary fonts are managed automatically
                                       #PB_Font_Bold, ; font_flag,
                                       #Black,        ; Text color
                                       0,             ; Text adjustment - 0=center - 1=left - 2=right
                                       -1,            ; Background color - Ignore = -1
                                       text$,         ; Text
                                       -5,            ; padding_x
                                       0,             ; padding_y
                                       1)             ; resize_factor.f
            PresetMaxFontsize_FW()
            
            If menu_ID
              DisableMenuItem(menu_ID, 3, 0)
            EndIf
          Else
            ; TODO Font_Wizzard_BF enhancement
            PresetMaxFontsize_FW(11)
            AddFrame_FW(1, $FF000000)
            Define text$="ERROR"+#LF$+"Can not create the KeyFile"
            SetImageBasedGadgetText_FW(button_state_ID,
                                       "",            ; Font$, ; The necessary fonts are managed automatically
                                       #PB_Font_Bold, ; font_flag,
                                       #Black,        ; Text color
                                       0,             ; Text adjustment - 0=center - 1=left - 2=right
                                       -1,            ; Background color - Ignore = -1
                                       text$,         ; Text
                                       -5,            ; padding_x
                                       0,             ; padding_y
                                       1)             ; resize_factor.f
            PresetMaxFontsize_FW()
          EndIf
        EndIf
      EndIf
      
      If EventMenu()=3
        Define rnd.q=CryptRandomData(@rnd, 8)
        QAES_Rabbit_SetCryptExtender("")
        Define i
        For i=1 To 5
          QAES_Rabbit_FileEncrypter(KeyFile$, password$+Str(rnd)+Str(i), 0, -1)
        Next
        rnd=0
        QAES_Rabbit_SetCryptExtender()
        If DeleteFile(KeyFile$)
          ; TODO Font_Wizzard_BF enhancement
          PresetMaxFontsize_FW(11)
          AddFrame_FW(1, $FF000000)
          Define text$="KeyFile deleted"
          SetImageBasedGadgetText_FW(button_state_ID,
                                     "",            ; Font$, ; The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     -1,            ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          PresetMaxFontsize_FW()
          
          If menu_ID
            DisableMenuItem(menu_ID, 3, 1)
          EndIf
        Else
          ; TODO Font_Wizzard_BF enhancement
          PresetMaxFontsize_FW(11)
          AddFrame_FW(1, $FF000000)
          Define text$="ERROR"+#LF$+"Can not delete the KeyFile"
          SetImageBasedGadgetText_FW(button_state_ID,
                                     "",            ; Font$, ; The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     -1,            ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          PresetMaxFontsize_FW()
        EndIf
      EndIf
      
      If EventMenu()=4
        Define keyFile_1$=OpenFileRequester(" Select a KeyFile for loading ", "" ,"" ,0)
        ; TODO Font_Wizzard_BF enhancement
        PresetMaxFontsize_FW(11)
        AddFrame_FW(1, $FF000000)
        Define text$=""
        SetImageBasedGadgetText_FW(button_state_ID,
                                   "",            ; Font$, ; The necessary fonts are managed automatically
                                   #PB_Font_Bold, ; font_flag,
                                   #Black,        ; Text color
                                   0,             ; Text adjustment - 0=center - 1=left - 2=right
                                   -1,            ; Background color - Ignore = -1
                                   text$,         ; Text
                                   -5,            ; padding_x
                                   0,             ; padding_y
                                   1)             ; resize_factor.f
        PresetMaxFontsize_FW()
        Define old_password$=password$
        If keyFile_1$<>"" And FileSize(keyFile_1$)=64
          Define file=ReadFile(#PB_Any, KeyFile_1$)
          password$=""
          If file
            password$=ReadString(file, #PB_Ascii)
            CloseFile(file)
          EndIf
          If Len(password$)=64
            old_password$=""
            ; TODO Font_Wizzard_BF enhancement
            PresetMaxFontsize_FW(11)
            AddFrame_FW(1, $FF000000)
            Define text$="KeyFile is imported"
            SetImageBasedGadgetText_FW(button_state_ID,
                                       "",            ; Font$, ; The necessary fonts are managed automatically
                                       #PB_Font_Bold, ; font_flag,
                                       #Black,        ; Text color
                                       0,             ; Text adjustment - 0=center - 1=left - 2=right
                                       -1,            ; Background color - Ignore = -1
                                       text$,         ; Text
                                       -5,            ; padding_x
                                       0,             ; padding_y
                                       1)             ; resize_factor.f
            PresetMaxFontsize_FW()
            
            If menu_ID
              DisableMenuItem(menu_ID, 3, 0)
            EndIf
          Else
            password$=old_password$
          EndIf
        EndIf
        If old_password$<>"" And keyFile_1$<>""
          ; TODO Font_Wizzard_BF enhancement
          PresetMaxFontsize_FW(11)
          AddFrame_FW(1, $FF000000)
          Define text$="ERROR"+#LF$+"KeyFile not imported"
          SetImageBasedGadgetText_FW(button_state_ID,
                                     "",            ; Font$, ; The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     -1,            ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          PresetMaxFontsize_FW()
        EndIf
        old_password$=""
      EndIf
      
      If EventMenu()=5
        Define keyFile_1$=SaveFileRequester(" Export a KeyFile", "Sakis_3Rabbit_KeyFile" ,"" ,0)
        ; TODO Font_Wizzard_BF enhancement
        PresetMaxFontsize_FW(11)
        AddFrame_FW(1, $FF000000)
        Define text$=""
        SetImageBasedGadgetText_FW(button_state_ID,
                                   "",            ; Font$, ; The necessary fonts are managed automatically
                                   #PB_Font_Bold, ; font_flag,
                                   #Black,        ; Text color
                                   0,             ; Text adjustment - 0=center - 1=left - 2=right
                                   -1,            ; Background color - Ignore = -1
                                   text$,         ; Text
                                   -5,            ; padding_x
                                   0,             ; padding_y
                                   1)             ; resize_factor.f
        PresetMaxFontsize_FW()
        
        file=CreateFile(#PB_Any, KeyFile_1$)
        If IsFile(file)
          If WriteString(file, password$, #PB_Ascii)=Len(password$)
            CloseFile(file)
            ; TODO Font_Wizzard_BF enhancement
            PresetMaxFontsize_FW(11)
            AddFrame_FW(1, $FF000000)
            Define text$="KeyFile exported"
            SetImageBasedGadgetText_FW(button_state_ID,
                                       "",            ; Font$, ; The necessary fonts are managed automatically
                                       #PB_Font_Bold, ; font_flag,
                                       #Black,        ; Text color
                                       0,             ; Text adjustment - 0=center - 1=left - 2=right
                                       -1,            ; Background color - Ignore = -1
                                       text$,         ; Text
                                       -5,            ; padding_x
                                       0,             ; padding_y
                                       1)             ; resize_factor.f
            PresetMaxFontsize_FW()
            
            If menu_ID
              DisableMenuItem(menu_ID, 3, 0)
            EndIf
          Else
            ; TODO Font_Wizzard_BF enhancement
            PresetMaxFontsize_FW(11)
            AddFrame_FW(1, $FF000000)
            Define text$="ERROR"+#LF$+"Can not export the KeyFile"
            SetImageBasedGadgetText_FW(button_state_ID,
                                       "",            ; Font$, ; The necessary fonts are managed automatically
                                       #PB_Font_Bold, ; font_flag,
                                       #Black,        ; Text color
                                       0,             ; Text adjustment - 0=center - 1=left - 2=right
                                       -1,            ; Background color - Ignore = -1
                                       text$,         ; Text
                                       -5,            ; padding_x
                                       0,             ; padding_y
                                       1)             ; resize_factor.f
            PresetMaxFontsize_FW()
          EndIf
        EndIf
      EndIf
      
      If EventMenu()=6
        Define file$=SaveFileRequester(" Delete and shredder a File", "" ,"" ,0)
        QAES_Rabbit_SetCryptExtender("")
        Define rnd.q=CryptRandomData(@rnd, 8)
        Define i
        For i=1 To 5
          QAES_Rabbit_FileEncrypter(file$, password$+Str(rnd)+Str(i), 0, progressbar_ID)
        Next 
        rnd=0
        QAES_Rabbit_SetCryptExtender()
        SetGadgetState(progressbar_ID, 0)
        Progressbar_EX(progressbar_ID, 0, 0, $FFFFFF) ; Tweak the OS progressbar :)
        If DeleteFile(file$)
          ; TODO Font_Wizzard_BF enhancement
          PresetMaxFontsize_FW(11)
          AddFrame_FW(1, $FF000000)
          Define text$="File deleted"
          SetImageBasedGadgetText_FW(button_state_ID,
                                     "",            ; Font$, ; The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     -1,            ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          PresetMaxFontsize_FW()
          
          If menu_ID
            DisableMenuItem(menu_ID, 3, 1)
          EndIf
        Else
          ; TODO Font_Wizzard_BF enhancement
          PresetMaxFontsize_FW(11)
          AddFrame_FW(1, $FF000000)
          Define text$="ERROR"+#LF$+"Can not delete the File"
          SetImageBasedGadgetText_FW(button_state_ID,
                                     "",            ; Font$, ; The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     -1,            ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          PresetMaxFontsize_FW()
        EndIf
      EndIf
      
      If EventMenu()=7
        MessageRequester("Hints",
                         "Use on your own risk, no warranty whats ever"+#LF$+#LF$+
                         "File crypter and UI by Saki and walbus"+#LF$+#LF$+
                         "ASM Rabbit crypter by Wilbert"+#LF$+#LF$+
                         "Delete KeyFiles and files ever with the crypter functions,"+#LF$+
                         "this destroys the Files before deleting them"+#LF$+#LF$+
                         "Special three-stage crypter with CBC mode based IV handling")   
      EndIf
      
      Delay(300)
      While WindowEvent() : Wend  
    EndIf
    
    Define event_gadget=EventGadget()
    Select event_gadget
      Case button_encrypt_ID, button_decrypt_ID
        DisableGadget(button_break_ID, 0)
        If event_gadget=button_encrypt_ID
          SetGadgetData(button_encrypt_ID, 1)
        Else
          SetGadgetData(button_encrypt_ID, 0)
        EndIf
        If win_event=#PB_Event_GadgetDrop
          If event_gadget=button_encrypt_ID
            DisableGadget(button_encrypt_ID, 1)
          Else
            DisableGadget(button_decrypt_ID, 1)
          EndIf
          progress_1
          If event_gadget=button_encrypt_ID
            DisableGadget(button_encrypt_ID, 0)
          Else
            DisableGadget(button_decrypt_ID, 0)
          EndIf
          Delay(200)
          SetGadgetState(progressbar_ID, 0)
          Progressbar_EX(progressbar_ID, 0, 0, $FFFFFF) ; Tweak the OS progressbar :)
        Else
          If event_gadget=button_encrypt_ID
            path$=OpenFileRequester("Select a file to encrypting !", "", "*.*", 0)
          Else
            path$=OpenFileRequester("Select a file to decrypting !", "", "*.*", 0)
          EndIf
        EndIf
        If path$<>""
          ; TODO Font_Wizzard_BF enhancement
          PresetMaxFontsize_FW(11)
          AddFrame_FW(1, $FF000000)
          Define text$=QAES_Rabbit_FileCrypter(path$, password$, integrity_check, progressbar_ID)
          SetImageBasedGadgetText_FW(button_state_ID,
                                     "",            ; Font$, ; The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     -1,            ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          PresetMaxFontsize_FW()
          
          Delay(200)
          SetGadgetState(progressbar_ID, 0)
          Progressbar_EX(progressbar_ID, 0, 0, $FFFFFF) ; Tweak the OS progressbar :)
        EndIf
        DisableGadget(button_break_ID, 1)
        While WindowEvent() : Wend
        
      Case button_integrity_check_ID
        DisableGadget(button_break_ID, 0)
        If win_event=#PB_Event_GadgetDrop
          cryptfile_check=1
          DisableGadget(button_integrity_check_ID, 1)
          progress_2
          DisableGadget(button_integrity_check_ID, 0)
          cryptfile_check=0
          Delay(200)
          SetGadgetState(progressbar_ID, 0)
          Progressbar_EX(progressbar_ID, 0, 0, $FFFFFF) ; Tweak the OS progressbar :)
        Else
          path$=OpenFileRequester("Select a encrypted file to integrity check !", "", "*.*", 0)
        EndIf
        If path$<>""
          ; TODO Font_Wizzard_BF enhancement
          PresetMaxFontsize_FW(11)
          AddFrame_FW(1, $FF000000)
          Define text$=QAES_Rabbit_CryptedFileCheck(path$, password$, progressbar_ID)
          SetImageBasedGadgetText_FW(button_state_ID,
                                     "",            ; Font$, ; The necessary fonts are managed automatically
                                     #PB_Font_Bold, ; font_flag,
                                     #Black,        ; Text color
                                     0,             ; Text adjustment - 0=center - 1=left - 2=right
                                     -1,            ; Background color - Ignore = -1
                                     text$,         ; Text
                                     -5,            ; padding_x
                                     0,             ; padding_y
                                     1)             ; resize_factor.f
          PresetMaxFontsize_FW()
          
          Delay(200)
          SetGadgetState(progressbar_ID, 0)
          Progressbar_EX(progressbar_ID, 0, 0, $FFFFFF) ; Tweak the OS progressbar :)
        EndIf
        DisableGadget(button_break_ID, 1)
        While WindowEvent() : Wend
    EndSelect
  Until win_event=#PB_Event_CloseWindow
  
  ; TODO Font_Wizzard_BF enhancement
  FreeAllGadgets_FW()
  
CompilerEndIf

; IDE Options = PureBasic 5.73 LTS (Windows - x64)
; CursorPosition = 262
; FirstLine = 259
; Folding = --------
; EnableXP
; DPIAware
; Executable = ..\3Rabbit.exe